class StateMachineIo:
    pass

class Connection:
    def __init__(self, o, t, i):
        self.o = o
        self.t = t
        self.i = i

class StateMachine:
    def __init__(self):
        self.i = StateMachineIo()
        self.s = StateMachineIo()
        self.o = StateMachineIo()
        self.con = []
        def f(_self):
            pass
        self._step = f
    def step(self):
        self._step(self)
    def send(self):
        for con in self.con:
            con.t.i.__dict__[con.i] = self.o.__dict__[con.o]
    def connect(self, _o_name : str, target, _i_name : str):
        self.con.append(Connection(_o_name, target, _i_name))
    def inner(self):
        def deco(f):
            def f2(_self):
                f(_self.i, _self.s, _self.o)
            self._step = f2
            return f2
        return deco
    def __str__(self):
        return ' '.join([l+':'+(','.join(['{}={}'.format(k,v) for k, v in t.__dict__.items()]) if len(t.__dict__)>0 else 'None')for t,l in [(self.i,'in'),(self.s,'self'),(self.o,'out')]])
            


from PIL import Image, ImageDraw, ImageFont
from .machine import StateMachine, StateMachineIo
from .test import gen_states, test_machine
import math

def gen_graph(machine : StateMachine, i : StateMachineIo, s : StateMachineIo, font)->Image:
    res = test_machine(machine, i, s)
    state_count = len(res.states)
    angle = 6.28/state_count
    size = state_count*128
    img = Image.new('RGB', (int(1.618*size), size), color='white')
    draw = ImageDraw.Draw(img)

    _font = ImageFont.truetype(font, 16)

    r = 32
    for i in range(state_count):
        pos_x = math.sin(i*angle)*size/3+size/2
        pos_y = math.cos(i*angle)*size/3+size/2
        draw.ellipse([pos_x-r,pos_y-r,pos_x+r,pos_y+r],outline='black')
        text = ','.join([str(k) for k in list(res.states[i])])
        bbox = draw.textbbox([pos_x, pos_y], text, font=_font)
        draw.text([bbox[0]-(bbox[2]-bbox[0])/2,bbox[1]-(bbox[3]-bbox[1])/2], text, font=_font, fill='black')
    
    lines = {}
    for i in range(len(res.data)):
        s1 = res.data[i][0]
        s2 = res.data[i][2]
        info = ','.join([str(k) for k in res.data[i][1]])+'/'+','.join([str(k) for k in res.data[i][3]])
        if not (s1,s2) in lines.keys():
            lines[(s1,s2)] = [info]
        else:
            lines[(s1,s2)].append(info)

    for k, v in lines.items():
        s1 = k[0]
        s2 = k[1]
        pos1_x = math.sin(s1*angle)*size/3+size/2
        pos1_y = math.cos(s1*angle)*size/3+size/2
        pos2_x = math.sin(s2*angle)*size/3+size/2
        pos2_y = math.cos(s2*angle)*size/3+size/2
        dx = pos2_x - pos1_x
        dy = pos2_y - pos1_y
        dl = math.sqrt(dx*dx+dy*dy)
        if s1!=s2:
            ix = dx/dl
            iy = dy/dl
            lx = -iy
            ly = ix
            draw.line([pos1_x+ix*r,pos1_y+iy*r,pos2_x-ix*r,pos2_y-iy*r],fill='black')
            draw.line([pos2_x-ix*r,pos2_y-iy*r,pos2_x-ix*(r+4)+lx*3,pos2_y-iy*(r+4)+ly*3],fill='black')
            draw.line([pos2_x-ix*r,pos2_y-iy*r,pos2_x-ix*(r+4)-lx*3,pos2_y-iy*(r+4)-ly*3],fill='black')
            tx = pos1_x+ix*72
            ty = pos1_y+iy*72
        else:
            draw.ellipse([pos1_x-r/2+r*math.sin(s1*angle),pos1_y-r/2+r*math.cos(s1*angle),pos1_x+r/2+r*math.sin(s1*angle),pos1_y+r/2+r*math.cos(s1*angle)], outline='black')  
            tx = pos1_x+2*r*math.sin(s1*angle)
            ty = pos1_y+2*r*math.cos(s1*angle)
        draw.multiline_text([tx,ty], '\n'.join(v),fill='black',font=_font)


    draw.text([0,0], 'state: '+','.join(res.k2)+'\tinput: '+','.join(res.k1)+'\toutput: '+','.join(res.k3),font=_font,fill='red')
    return img



from .machine import StateMachine, StateMachineIo

def gen_states(s : StateMachineIo):
    states = []
    keys = list(s.__dict__.keys())
    if len(keys)>0:
        states = [[v] for v in s.__dict__[keys[0]]]
        if len(keys)>1:
            for k in keys[1:]:
                new_states = []
                for i in states:
                    for j in s.__dict__[k]:
                        new_states.append(i+[j])
                states = new_states        
    states = [tuple(item) for item in states]
    back = {}
    for i in range(len(states)):
        back[states[i]]=i
    return states, keys, back

class TestResult:
    def __init__(self,k1,k2,k3,data,states):
        self.k1 = k1
        self.k2 = k2
        self.k3 = k3
        self.data = data
        self.states = states

def test_machine(machine : StateMachine, i : StateMachineIo, s : StateMachineIo):
    inputs, i_keys, i_dict = gen_states(i)
    states, s_keys, s_dict = gen_states(s)
    
    o_keys = machine.o.__dict__.keys()

    data = []
    for i in range(len(states)):
        for j in range(len(inputs)):
            m = StateMachine()
            for k in range(len(s_keys)):
                m.s.__dict__[s_keys[k]] = states[i][k]
            for k in range(len(i_keys)):
                m.i.__dict__[i_keys[k]] = inputs[j][k]
            for k, v in machine.o.__dict__.items():
                m.o.__dict__[k] = v
            machine._step(m)
            res_s = tuple([m.s.__dict__[k] for k in s_keys])
            res_o = tuple([m.o.__dict__[k] for k in o_keys])
            data.append([s_dict[states[i]],inputs[j],s_dict[res_s], res_o])

    return TestResult(i_keys, s_keys, o_keys, data, states)
    